#
#    ========== licence begin LGPL
#    Copyright (C) 2002 SAP AG
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    ========== licence end
#

import sys
import string
import os

import fileutil

from SchemaCompiler import SchemaCompiler

class ExpatC (SchemaCompiler):
    schemaName = '<unknown>'
    def defaultName (self, base):
        self.schemaName = base
        return base + 'Parser.cpp'

    def parserClass (self):
        return self.schemaName + 'Parser'

    def writeIntro (self, stream, fname):
        # self.writeHeaderFile (fname)
        stream.write (cformat ('', (fname, self.schemaName), [
            '/*',
            '    module:     %s',
            '    generated from: %s',
            '*/',
            '',
            '#include "ZExternal/expat/expat.h"',
            '#include "XML/FOMParser/FOM_IElementHandler.hpp"',
            '#include "XML/FOMParser/FOM_AbstractParser.hpp"',
            ]))
        hppIndex = 0
        while 1:
            header = self.schema.getProperty ('hpp.%d' % hppIndex)
            if not header:
                break
            stream.write ('#include "%s"\n' % header)
            hppIndex = hppIndex + 1
        for element in self.schema.elements:
            if element.isGroup:
                continue
            includeFile = element.getProperty ('includefile')
            if not includeFile:
                includePattern = element.getProperty ('includepattern')
                if includePattern:
                    cppName = self.getElementClassName (element)
                    includeFile = includePattern % cppName
            if includeFile:
                stream.write ('#include "%s"\n' % includeFile)

    def writeHeaderFile (self, cppName):
        schemaName = self.schemaName
        fname = schemaName + 'Parser.hpp'
        stream = open (fname, 'w')
        includeGuard = string.upper (schemaName) + 'PARSER_HPP'
        parserClass = self.parserClass ()
        stream.write (cformat ('', locals (), [
            '/*',
            '    module:     %(fname)s',
            '    generated from: %(schemaName)s',
            '*/',
            '',
            '#ifndef %(includeGuard)s',
            '#define %(includeGuard)s',
            '',
            '#include "XML/FOMParser/FOM_AbstractParser.hpp"',
            '',
            'class %(parserClass)s : public FOM_AbstractParser',
            '{',
            '  public:',
            '    virtual const FOM_IElementHandler * findTagHandler (const char * name);',
            '};',
            ]))
        stream.write ('\n#endif\n')
        stream.close ()


    def compileElement (self, element, stream):
        if element.isGroup:
            return
        handlerClass = self.getHandlerClassName (element)
        self.writeMethodFluff (stream)
        stream.write (cformat ('', locals (), [
            'class %(handlerClass)s: public FOM_IElementHandler',
            '{',
            'public:',
            '    virtual void * create (const char ** atts) const;',
            #'    virtual void   setAttribute (void * element,',
            #'                       const char * attName,',
            #'                       const char * attValue) const;',
            '    virtual void   addElement (void * element,',
            '                       const char * name,',
            '                       void * subElement) const;',
            '    virtual void   setCharacterData (void * element,',
            '                       const string & data) const;',
            '    static const %(handlerClass)s instance;',
            '};',
            '',
            'const %(handlerClass)s %(handlerClass)s::instance;',
            ]))
        for name in ['create', 'addElement', 'setCharacterData']:
            try:
                method = getattr (self, 'write_' + name)
                self.writeMethodFluff (stream)
                method (element, stream)
            except AttributeError:
                pass

    def writeFactory (self, stream):
        stream.write (cformat ('', None, [
            '/*----------------------------------------------------------------------',
            '',
            '  chapter: handler factory',
            '',
            '----------------------------------------------------------------------*/',
            '',
            ]))
        #
        # write lookup table
        #
        stream.write (cformat ('', locals (), [
            'static const struct {',
            '    const char * tagName;',
            '    const FOM_IElementHandler * handler;',
            '} tagMapping [] = {',
            ]))
        lookup = []
        for element in self.schema.elements:
            if element.isGroup:
                continue
            if element.getProperty ('isString'):
                handlerClass = 'FOM_PurePCDataHandler'
            else:
                handlerClass = self.getHandlerClassName (element)
            lookup.append ((element.name, handlerClass))
        lookup.sort ()
        for name, handlerClass in lookup:
            stream.write ('    {"%s", &%s::instance},\n' % (name, handlerClass))
        stream.write (cformat ('', None, [
            '    {NULL, NULL}',
            '};',
            ]))
        #
        # write lookup method
        #
        #self.writeMethodFluff (stream)
        #parserClass = self.parserClass ()
        #stream.write (cformat ('', locals (), [
        #    'class %(parserClass)s : public FOM_AbstractParser',
        #    '{',
        #    '  public:',
        #    '    virtual const FOM_IElementHandler * findTagHandler (const char * name);',
        #    '};',
        #    ]))
        self.writeMethodFluff (stream)
        schemaClass = self.schemaName + 'Parser'
        elementCount = len (lookup)
        stream.write (cformat ('', locals (), [
            'const FOM_IElementHandler * %(schemaClass)s::findTagHandler (',
            '    const char * name)',
            '{',
            '    int low = 0;',
            '    int current;',
            '    int high = %(elementCount)s;',
            '    int compareResult;',
            '    while ((current != low) && (current != high)) {',
            '        current = (high - low) / 2 + low;',
            '        compareResult = strcmp (name, tagMapping [current].tagName);',
            '        if (compareResult == 0) {',
            '            return tagMapping [current].handler;',
            '        }',
            '        if (compareResult < 0) {',
            '            high = current;',
            '        }',
            '        else {',
            '            low = current;',
            '        }',
            '    }',
            '    return NULL;',
            '}',
            ]))

    def getHandlerClassName (self, element):
        prefix = element.getProperty ('handlerclassprefix', '')
        return prefix + element.name + 'Handler'

    def getElementClassName (self, element):
        prefix = element.getProperty ('realclassprefix', '')
        return prefix + element.name

    def getInterfaceClassName (self, element):
        groupName = element.getProperty ('group')
        if not groupName:
            return self.getElementClassName (element)
        superElement = self.schema.get (groupName)
        return self.getElementClassName (superElement)

    def expandSubclasses (self, namelist, isListEntry):
        result = []
        for name in namelist:
            element = self.schema.get (name)
            elements = element.getProperty ('elements')
            if elements:
                for subClass in elements:
                    result.append ((name, subClass.name, isListEntry))
            else:
                result.append ((name, name, isListEntry))
        return result

    def writeMethodFluff (self, stream):
        stream.write ('\n/*----------------------------------------*/\n\n')

    def write_create (self, element, stream):
        handlerClass = self.getHandlerClassName (element)
        cppClass = self.getElementClassName (element)
        stream.write (cformat ('', locals (), [
            'void  * %(handlerClass)s::create (',
            '    const char ** atts) const',
            '{',
            ]))
        #
        # declare attributes
        #
        if element.attributes is None:
            attributes = []
        else:
            attributes = element.attributes
        def varName (attribute):
            return attribute.name + 'Variable'

        for attribute in attributes:
            variableName = varName (attribute)
            dataType = attribute.getProperty ('datatype')
            default = attribute.default ()
            defaultValue = ''
            if dataType:
                typeString = dataType
                if default:
                    defaultValue = ' = %s' % default
            else:
                typeString = 'const char *'
                if default:
                    defaultValue = ' = "%s"' % default
                else:
                    defaultValue = ' = 0'
            stream.write (cformat ('    ', locals (), [
                '%(typeString)s %(variableName)s%(defaultValue)s;',
                ]))
        #
        # parse attributes
        #
        if attributes:
            stream.write (cformat('    ', None, [
                'for (int attsIndex = 0; atts [attsIndex] != NULL; attsIndex += 2) {',
                '    const char * attName = atts [attsIndex];',
                '    const char * attValue = atts [attsIndex + 1];'
                ]))
            delimiter = ''
            for attribute in attributes:
                attName = attribute.name
                variableName = varName (attribute)
                convExpression = 'attValue'
                dataType = attribute.getProperty ('datatype')
                if dataType:
                    conversion = self.schema.getProperty ('parseconversion.' + dataType)
                    if conversion is None:
                        conversion = 'dont_know_how_to_convert_chars_to_' + dataType
                    convExpression = conversion + ' (attValue)'
                stream.write (cformat ('        ', locals (), [
                    'if (strcmp (attName, "%(attName)s") == 0) {',
                    '    %(variableName)s = %(convExpression)s;',
                    '}',
                    ]))
                delimiter = 'else '
            stream.write (cformat ('        ', None, [
                'else {',
                '    /* TODO: throw exception because of unknown attribute */',
                '}',
                ]))
            stream.write ('    }\n')
        if not attributes:
            stream.write (cformat  ('    ', None, [
                'if (atts [0] != NULL) {',
                '    /* TODO: throw exception because of unknown attribute */',
                '}',
                ]))
        #
        # create new object
        #
        constructorArgs = []
        for attribute in attributes:
            if attribute.getProperty ('constructorparameter'):
                constructorArgs.append (varName (attribute))
        argumentList = string.join (constructorArgs, ', ')
        allocator = self.schema.getProperty ('allocator', '')
        if allocator:
            allocator = '(' + allocator + ')'
        stream.write (cformat ('    ', locals (), [
            '%(cppClass)s * result = new %(allocator)s %(cppClass)s (%(argumentList)s);',
            ]))
        # postnew
        snippetId = element.getProperty ('postnew')
        if snippetId:
            snippet = self.schema.getProperty ('customcode.' + snippetId)
            if snippet:
                code = snippet
            else:
                code = snippetId
            stream.write ('    ' + code + '\n')
        #
        # set non constructor arguments
        #
        for attribute in attributes:
            if not attribute.getProperty ('constructorparameter'):
                methodName = attribute.getProperty ('setter', 'set' + attribute.name)
                variableName = attribute.name + 'Variable'
                stream.write (cformat ('    ', locals (), [
                    'result->%(methodName)s (%(variableName)s);',
                    ]))
        #
        # finish method
        #
        stream.write (cformat ('', None, [
            '    return result;',
            '}',
            ]))

    def write_setAttribute (self, element, stream):
        handlerClass = self.getHandlerClassName (element)
        cppClass = self.getElementClassName (element)
        stream.write (cformat ('', locals (), [
            'void %(handlerClass)s::setAttribute (',
            '    void         * element,',
            '    const char   * attName,',
            '    const char   * attValue) const',
            '{',
            '    %(cppClass)s * self = reinterpret_cast<%(cppClass)s *> (element);',
            ]))
        if element.attributes:
            delimiter = ''
            for attr in element.attributes:
                attrName = attr.name
                variableName, conversionLine = self.generateParseConversion (
                    attr, 'attValue')
                methodName = attr.getProperty ('setter', 'set' + attrName)
                stream.write (cformat ('    ', locals (), [
                    delimiter + 'if (strcmp (attName, "%(attrName)s") == 0) {',
                    conversionLine,
                    '    self->%(methodName)s (%(variableName)s);',
                    '}',
                    ]))
                delimiter = 'else '
            stream.write (cformat ('    ', None, [
                'else {',
                '    /* TODO: throw exception because of unknown attribute */',
                '}'
                ]))
        else:
            stream.write ("    /* TODO: throw exception because of unknown attribute */\n")
        stream.write ("\n}\n")

    def write_addElement (self, element, stream):
        handlerClass = self.getHandlerClassName (element)
        cppClass = self.getElementClassName (element)
        stream.write (cformat ('', locals (), [
            'void %(handlerClass)s::addElement (',
            '    void         * element,',
            '    const char   * name,',
            '    void         * subElement) const',
            '{',
            '    %(cppClass)s * self = reinterpret_cast<%(cppClass)s *> (element);',
            ]))
        varnames = self.expandSubclasses (element.varnames (), None)
        elementClass = self.getElementClassName (element)
        listnames = self.expandSubclasses (element.listnames (), 'true')
        allnames = varnames + listnames
        if allnames:
            delimiter = ''
            for varname, subClassName, isListEntry in allnames:
                if isListEntry:
                    prefix = 'add'
                else:
                    prefix = 'set'
                methodName = element.getProperty (varname + '.setter', prefix + varname)
                subElement = self.schema.elementDict [subClassName]
                if subElement.getProperty ('isString'):
                    subElementClass = 'FOM_PurePCData'
                    expr = 'typedElement->getData ()'
                    delete = '    delete typedElement;'
                else:
                    subElementClass = self.getElementClassName (subElement)
                    expr = 'typedElement'
                    delete = ''
                stream.write (cformat ('    ', locals (), [
                    delimiter + 'if (strcmp (name, "%(subClassName)s") == 0) {',
                    '    %(subElementClass)s * typedElement = reinterpret_cast<%(subElementClass)s *> (subElement);',
                    '    self->%(methodName)s (%(expr)s);',
                    delete,
                    '}',
                    ]))
                delimiter = 'else '
            stream.write (cformat ('    ', None, [
            'else {',
            '    /* TODO: throw exception because of unknown sub element */',
            '}',
            ]))
        else:
            stream.write ("    /* TODO: throw exception because of unknown subelement */\n")
        stream.write ("\n}\n")

    def write_setCharacterData (self, element, stream):
        handlerClass = self.getHandlerClassName (element)
        cppClass = self.getElementClassName (element)
        stream.write (cformat ('', locals (), [
            'void %(handlerClass)s::setCharacterData (',
            '    void         * element,',
            '    const string & data) const',
            '{',
            '    %(cppClass)s * self = reinterpret_cast<%(cppClass)s *> (element);',
            ]))
        stream.write ("""    /* TODO: convert string to other datatypes */\n""")
        stream.write ("""    /* TODO: get set method*/\n""")
        stream.write ("\n}\n")

    def generateParseConversion (self, item, varname):
        variableName = varname
        dataType = item.getProperty ('datatype')
        if dataType:
            variableName = 'typedValue'
            conversion = self.schema.getProperty ('parseconversion.' + dataType)
            if conversion is None:
                conversion = 'dont_know_how_to_convert_chars_to_' + dataType
            conversionLine = ('    %(dataType)s %(variableName)s '
                + '= %(conversion)s (%(varname)s);') % locals ()
        else:
            conversionLine = '    /* no conversion */'
        return variableName, conversionLine

def cformat (indent, formatArg, lines):
    text = indent + string.join (lines, '\n' + indent) + '\n'
    if formatArg:
        text = text % formatArg
    return text

TODOisString = """
- generate dummy class
- use name of dummy class (getElementClassName)
- set object.stringValue () instead of object
- delete object
"""

